/* 3 plane grayscales by Kevin Kofler - Kevin@tigcc.ticalc.org */

#ifndef __GRAY3P
#define __GRAY3P

#include <default.h>

/*===========================================================================*/
/*                                                                           */
/* GrayScale-Support v3.50-3P for TIGCC                                      */
/*                                                                           */
/* compatible with HW1/HW2 on all AMS versions up to 2.05                    */
/*                                                                           */
/*                                                                           */
/* original implementation:       Julien Muchembled (as part of UniversalOS) */
/* NOSTUB and OpenSource version: thomas.nussbaumer@gmx.net (TiCT)           */
/* 3 plane (7/8 grayscales) version: Kevin@tigcc.ticalc.org (TIGCC Team)     */
/*                                                                           */
/*===========================================================================*/

/*--------------------------------------------------------------------------*/
/* INTERNAL SECTION                                                         */
/*                                                                          */
/* DON'T USE THE INTERNAL VARIABLES DIRECTLY - INSTEAD USE THE MACROS BELOW */
/*--------------------------------------------------------------------------*/

extern void*                   __gray3P_L_plane;     // pointer to light  plane of standard buffer
extern void*                   __gray3P_M_plane;     // pointer to medium plane of standard buffer
extern void*                   __gray3P_D_plane;     // pointer to dark   plane of standard buffer
extern void*                   __gray3P_L_plane2;    // pointer to light  plane of dbuffer
extern void*                   __gray3P_M_plane2;    // pointer to medium plane of dbuffer
extern void*                   __gray3P_D_plane2;    // pointer to dark   plane of dbuffer
extern unsigned short          __gray3P_handle;
extern short                   __gray3P_hw_type;
extern unsigned short          __gray3P_dbl_offset;  // offset to active grayscale doublebuffer (only values 0 and 12 are allowed)
extern volatile unsigned long  __gray3P_switch_cnt;
extern volatile void          *__gray3P_old_int1_hw1;// old INT1 handler on HW1
extern volatile void          *__gray3P_old_int1_hw2;// old INT1 handler on HW2
//extern const char              __gray3P_version[];   // just for internal maintainance

/*--------------------------------------------------------------------------*/
/* PUBLIC SECTION                                                           */
/*--------------------------------------------------------------------------*/

/* Begin NOT Auto-Generated Part - need to adapt the hsf files before ! */
#define GRAY3P_DBUFFER_SIZE 11528
#ifndef __HAVE_Bool
#define __HAVE_Bool
enum Bool{FALSE,TRUE};
#endif
//enum GrayModes{GRAY_OFF=0,GRAY_ON=1,GRAY_HW1=1,GRAY_HW2=1};
enum Gray3PPlanes{GRAY3P_LIGHT_PLANE=0,GRAY3P_MEDIUM_PLANE=1,GRAY3P_DARK_PLANE=2};
#ifndef __HAVE_INT_HANDLER
#define __HAVE_INT_HANDLER
typedef struct{short foo;}_DEREF_INT_HANDLER,*INT_HANDLER;
#endif
//#define GrayAdjust(x) ((void)(*(volatile unsigned char*)0x600013=128-(signed char)(x)))
#define Gray3PCheckRunning() (!!__gray3P_handle)
//#define IsGrayMode GrayCheckRunning
#define Gray3PDBufCleanup() ((void)(__gray3P_dbl_offset=0,__gray3P_D_plane2=__gray3P_D_plane,__gray3P_M_plane2=__gray3P_M_plane,__gray3P_L_plane2=__gray3P_L_plane))
#define Gray3PDBufGetActiveIdx() (!!__gray3P_dbl_offset)
#define Gray3PDBufGetActivePlane(x) (Gray3PDBufGetPlane(Gray3PDBufGetActiveIdx(),x))
#define Gray3PDBufGetHiddenIdx() (!__gray3P_dbl_offset)
#define Gray3PDBufGetHiddenPlane(x) (Gray3PDBufGetPlane(Gray3PDBufGetHiddenIdx(),x))
#define Gray3PDBufGetPlane(i,x) ((i)?(((void **)(&__gray3P_L_plane2))[(x)]):Gray3PGetPlane(x))
#define Gray3PDBufInit(p) ({void*aptr=(void*)((((long)p)+7)&0xfffffff8L);__gray3P_dbl_offset=0;__gray3P_D_plane2=aptr;__gray3P_M_plane2=aptr+3840;(void)(__gray3P_L_plane2=aptr+7680);})
#define Gray3PDBufSetActiveAMSPlane(x) Gray3PDBufSetAMSPlane(Gray3PDBufGetActiveIdx(),x)
#define Gray3PDBufSetActiveIdx(i) ((void)(__gray3P_dbl_offset=((i)?12:0)))
#define Gray3PDBufSetActiveIdxSync(i) ((void)(Gray3PWaitNSwitches(1),Gray3PDBufSetActiveIdx(i)))
#define Gray3PDBufSetAMSPlane(i,x) (_rom_call(void,(void*,long),1A2)(Gray3PDBufGetPlane(i,x),0xEF007F))
#define Gray3PDBufSetHiddenAMSPlane(x) Gray3PDBufSetAMSPlane(Gray3PDBufGetHiddenIdx(),x)
#define Gray3PDBufToggle() ((void)(__gray3P_dbl_offset=(__gray3P_dbl_offset?0:12)))
#define Gray3PDBufToggleSync() ({Gray3PWaitNSwitches(1);Gray3PDBufToggle();})
#define Gray3PGetInt1Handler() ((INT_HANDLER)(__gray3P_hw_type?__gray3P_old_int1_hw2:__gray3P_old_int1_hw1))
//#define GetGrayInt1Handler GrayGetInt1Handler
#define Gray3PGetPlane(x) (((void **)(&__gray3P_L_plane))[(x)])
//#define GetPlane GrayGetPlane
#define Gray3PGetSwitchCount() (__gray3P_switch_cnt)
//#define GetGraySwitchCount GrayGetSwitchCount
//#define Gray3PGetVersionString() ((const char*)__gray_version)
//#define GrayMode(x) ((x)?GrayOn():({GrayOff();(short)1;}))
extern void Gray3POff(void)__ATTR_LIB_ASM__;
extern short Gray3POn(register short gray8 asm("d0"))__ATTR_LIB_ASM__;
#define Gray3PSetAMSPlane(x) (_rom_call(void,(void*,long),1A2)(Gray3PGetPlane(x),0xEF007F))
//#define SetPlane GraySetAMSPlane
#define Gray3PSetInt1Handler(p) ((void)(__gray3P_hw_type?((INT_HANDLER)__gray3P_old_int1_hw2=(p)):((INT_HANDLER)__gray3P_old_int1_hw1=(p))))
//#define SetGrayInt1Handler GraySetInt1Handler
#define Gray3PSetSwitchCount(val) (__gray3P_switch_cnt=(val))
//#define SetGraySwitchCount GraySetSwitchCount
#define Gray3PWaitNSwitches(n) ({unsigned long __w=__gray3P_switch_cnt+(n);while(__gray3P_switch_cnt<__w);})
/* End NOT Auto-Generated Part */

#define _Gray3PIsRealHW2() (__gray3P_hw_type)

#endif
